% Example of Batch vs Recursive form of solving unconstrained optimal
% control problem
% ME 234(b), Spring 2022
% Anushri Dixit, Joel Burdick
%%
clear; close all;
%% Model Data
% Continuous-time Dynamics
% x = [z, zdot, x, xdot, ...]
g = 9.81; m = 0.65; l = 0.23; Ixx = 0.0075; Iyy = 0.0075; Izz = 0.013;
nx = 12; nu = 4;
Ac = [0 1 0 0 0 0 0 0 0 0 0 0; ...
      0 0 0 0 0 0 0 0 0 0 0 0; ...
      0 0 0 1 0 0 0 0 0 0 0 0; ...
      0 0 0 0 0 0 0 0 0 0 0 0; ...
      0 0 0 0 0 1 0 0 0 0 0 0; ...
      0 0 0 0 0 0 g 0 0 0 0 0; ...
      0 0 0 0 0 0 0 1 0 0 0 0; ...
      0 0 0 0 0 0 0 0 0 0 0 0; ...
      0 0 0 0 0 0 0 0 0 1 0 0; ...
      0 0 0 0 0 0 0 0 0 0 g 0; ...
      0 0 0 0 0 0 0 0 0 0 0 1; ...
      0 0 0 0 0 0 0 0 0 0 0 0];
  Bc = [0 0 0 0; 1/m 0 0 0; ...
        0 0 0 0; 0 1/Ixx 0 0; ...
        0 0 0 0; 0 0 0 0; ...
        0 0 0 0; 0 0 1/Iyy 0; ...
        0 0 0 0; 0 0 0 0; ...
        0 0 0 0; 0 0 0 1/Izz];

dt = 0.1; % Controller at 10 Hz
%  Exact discretization
E = expm(dt.*[Ac, Bc; zeros(nu, nx), zeros(nu, nu)]);
A = E(1:nx, 1:nx);
B = E(1:nx, nx+1:end);

% Euler discretization
Ad = (eye(nx) + Ac.*dt);
Bd = Bc.*dt;

% Optimal Control Parameters
P = 10*eye(nx);
Q = eye(nx);
R = eye(nu);
N = 40;

%
IC = [-2; zeros(nx-1,1)];

%% Batch Approach
barQ = [];
barR = [];
Sx = [];
Su = zeros(nx,N*nu);
for i = 1:N
    barQ = blkdiag(barQ, Q);
    barR = blkdiag(barR, R);
    Sx = vertcat(Sx, A^(i-1));
    Su = vertcat(Su, horzcat(A^(i-1)*B, Su(nx*(i-1)+1:nx*i, 1:end-nu)));
end
barQ = blkdiag(barQ, P);
Sx = vertcat(Sx, A^N);

controller_batch = -1*inv(Su'*barQ*Su + barR)*Su'*barQ*Sx*IC;
states_batch = [IC];
batch_i = [IC, zeros(nx, N)];
for i=1:N
    rng(116)
    states_batch(:,i+1) = A*states_batch(:,i)+B*controller_batch(nu*(i-1)+1:nu*i) + 0.01*(rand(nx,1)-0.5);
end
plot3(states_batch(5,  :), states_batch(9,  :), states_batch(1,  :), 'og')
set(gca, 'ZDir','reverse')
states_batch(1,  :)

%% Recursive Approach
recP = [zeros(nx, nx*N), P];
controller_rec = cell([N,1]);
Pk = P;
for i = N:-1:1
    controller_rec{i} = -1.*inv(B'*Pk*B + R)*B'*Pk*A;
    Pk = A'*Pk*A + Q - A'*Pk*B*inv(B'*Pk*B + R)*B'*Pk*A;
    recP(:, nx*(i-1)+1:nx*i) = Pk;
end
states_rec = [IC];
for i = 1:N
    rng(116)
    states_rec(:,i+1) = A*states_rec(:,i)+B*controller_rec{i}*states_rec(:,i) + 0.01*(rand(nx,1)-0.5);
end
hold on
plot3(states_rec(5,  :), states_rec(9,  :), states_rec(1,  :), '*b')
states_rec(1,  :)
legend('Batch Approach', 'Recursive Approach','Location','NorthEast')
xlabel('x')
ylabel('y')
zlabel('z')