% MPC computation for a simple 2D system (unstable).
% Requires YALMIP and mosek installation
% ME 234(b), Spring 2022
% Anushri Dixit, Joel Burdick
%%
yalmip('clear'); clear all; close all;
nx = 2; nu = 1; nw = 1;
A = [1 0;0 1.1]; 
B = [1; 0.5];
D = [0.01; 0.01];

% Optimal Control Parameters
Q1 = 1.*eye(nx);
Q2 = 100.*[3 -2; -2 4];
R = 1.*eye(nu);
Q=Q2;
N1 = 2;
N2 = 12;
N3 = 50;
N4 = 1000;
N= N2;
P = dare(A, B, Q, R);

% Batch disturbance matrix
epsilon = 0.001;
S_w = zeros(nx, N);
dist_i = zeros(nx, N);
for (i = 1:N)
    dist_i = horzcat(A^(i-1)*D, dist_i(:, 1:end-1));
    S_w = vertcat(S_w, dist_i);
end
% Disturbance - Normal (0,1)
Sigma = eye(N*nw,N*nw);
tighten = icdf('normal',1-epsilon,0,1).*vecnorm((S_w*Sigma)',2)';

% Initial condition (IC)
IC = [1;1];
% Written as a decision variable so that we can solve for an arbitrary IC
x0 = sdpvar(nx,1);
lin_pt = sdpvar(nx,1);

% Obstacle position
obs = [-1;0];
r = 0.5;

% Control Input as the decision variable
u = sdpvar(repmat(nu,1,N),repmat(1,1,N));

% Set up constraints
constraints = [];
objective = 0;
x = x0;
for k = 1:N
 x = A*x + B*u{k};
 objective = objective + x'*Q*x + u{k}'*R*u{k};
%  % Constraints: Without disturbances
%  constraints = [constraints, -1 <= u{k}<= 1, [-5;-5]<=x<=[5;5]];
%  constraints = [constraints; x'*(x0-lin_pt) >= lin_pt'*(x0-lin_pt)];
 % Constraints: With disturbances
 constraints = [constraints, -1 <= u{k}<= 1, ...
     [-5;-5]<=x-tighten(nx*k+1:(nx*(k+1))), ...
     x+tighten(nx*k+1:(nx*(k+1)))<=[5;5]];
 constraints = [constraints; x'*(x0-lin_pt) - ...
     tighten(nx*k+1:(nx*(k+1)))'.*norm(x0-lin_pt,2) >= lin_pt'*(x0-lin_pt)];
end

% Terminal cost P_\infty
objective = objective + x'*P*x;

% Find controller as a function of the IC
ops = sdpsettings('solver', 'mosek', 'verbose',1);
controller = optimizer(constraints,objective,ops,[x0;lin_pt],u{1});

appliedU = [];
x_ic = IC;
clf;
X = x_ic;
intersect = [0;0];

rng(234);
hold on
for i = 1:200
  display(['iteration: ', num2str(i)])
  dist_obs_center = sqrt((obs(1)-x_ic(1))^2 + (obs(2)-x_ic(2))^2);
  intersect = [obs(1) + r*(x_ic(1)-obs(1))/dist_obs_center; ...
      obs(2) + r*(x_ic(2)-obs(2))/dist_obs_center];
  U = controller{[x_ic; intersect]};
  % Update state: Without Disturbances
%   x_ic = A*x_ic + B*U(:,1);
%   % Update state: With disturbances
  x_ic = A*x_ic + B*U(:,1) + D*normrnd(0,1);
  X = [X x_ic];
  appliedU = [appliedU; U(1)];
end
plot(X(1,:), X(2,:), '*-b')
hold on
ezplot(@(x,y) (x-obs(1)).^2 + (y-obs(2)).^2 -r^2)
axis([-5 1.1 -1.5 1.5])
xlabel('x_1')
ylabel('x_2')